/* IDS POP3 - a small, fast, and efficient POP3 daemon
   Copyright (C) 1999 Jakob 'sparky' Kaivo <jkaivo@nodomainname.net>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "pop3d.h"

int
main (int argc, char **argv)
{
  struct group *gr;
  static int mode;
  int maxchildren = 10;

  /* First we want our group to be mail so we can access the spool */
  gr = getgrnam ("mail");
  if (gr == NULL)
    {
      perror ("Error getting group");
      exit (1);
    }
  setgid (gr->gr_gid);

  /* Set the signal handlers */
  signal (SIGINT, pop3_signal);
  signal (SIGQUIT, pop3_signal);
  signal (SIGILL, pop3_signal);
  signal (SIGBUS, pop3_signal);
  signal (SIGFPE, pop3_signal);
  signal (SIGSEGV, pop3_signal);
  signal (SIGTERM, pop3_signal);
  signal (SIGSTOP, pop3_signal);

  if (argc > 3)
    pop3_usage ();
  else if (argc > 1)
    {
      if (!strcmp (argv[1], "-i"))
	{
	  if (argc != 2)
	    pop3_usage ();
	  mode = INTERACTIVE;
	}
      else if (!strcmp (argv[1], "-d"))
	{
	  mode = DAEMON;
	  if (argc == 3)
	    maxchildren = atoi (argv[2]);
	  if (maxchildren < 5)
	    pop3_usage ();
	}
      else
	pop3_usage ();
    }

  if (mode == DAEMON)
    pop3_daemon_init ();

  /* change directories */
#ifdef MAILSPOOLHOME
  chdir ("/");
#else
  chdir (_PATH_MAILDIR);
#endif

  /* Set up for syslog */
  openlog ("ids-pop3d", LOG_PID, LOG_MAIL);

  umask (S_IROTH | S_IWOTH | S_IXOTH);	/* 007 */

  /* Actually run the daemon */
  switch (mode)
    {
    case DAEMON:
      pop3_daemon (maxchildren);
      break;
    case INTERACTIVE:
    default:
      pop3_mainloop (fileno (stdin), fileno (stdout));
      break;
    }

  /* Close the syslog connection and exit */
  closelog ();
  return OK;
}

/* Sets things up for daemon mode */

void
pop3_daemon_init (void)
{
  if (fork ())
    exit (0);			/* parent exits */
  setsid ();			/* become session leader */

  if (fork ())
    exit (0);			/* new parent exits */

  /* close inherited file descriptors */
  close (0);
  close (1);
  close (2);

  signal (SIGHUP, SIG_IGN);	/* ignore SIGHUP */
  signal (SIGCHLD, pop3_signal);	/* for forking */
}

/* The main part of the daemon. This function reads input from the client and
   executes the proper functions. Also handles the bulk of error reporting. */

int
pop3_mainloop (int infile, int outfile)
{
  int status = OK;
  char buf[260], *arg, *cmd;

  ifile = infile;
  ofile = fdopen (outfile, "w+");
  if (ofile == NULL)
    pop3_abquit (ERR_NO_OFILE);
  state = AUTHORIZATION;
  lockfile = NULL;
  curr_time = time (NULL);

  syslog (LOG_INFO, "Incoming connection opened");

  fflush (ofile);
  fprintf (ofile, "+OK POP3 " WELCOME "\r\n");

  while (state != UPDATE)
    {
      fflush (ofile);
      status = OK;
      if (read (ifile, buf, 260) < 1)
	pop3_abquit (ERR_DEAD_SOCK);
      cmd = pop3_cmd (buf);
      arg = pop3_args (buf);

      if (strlen (arg) > 255 || strlen (cmd) > 255)
	status = ERR_TOO_LONG;
      else if (strlen (cmd) > 4)
	status = ERR_BAD_CMD;
      else if (strncasecmp (cmd, "USER", 4) == 0)
	status = pop3_user (arg);
      else if (strncasecmp (cmd, "QUIT", 4) == 0)
	status = pop3_quit (arg);
      else if (strncasecmp (cmd, "APOP", 4) == 0)
	status = pop3_apop (arg);
      else if (strncasecmp (cmd, "AUTH", 4) == 0)
	status = pop3_auth (arg);
      else if (strncasecmp (cmd, "STAT", 4) == 0)
	status = pop3_stat (arg);
      else if (strncasecmp (cmd, "LIST", 4) == 0)
	status = pop3_list (arg);
      else if (strncasecmp (cmd, "RETR", 4) == 0)
	status = pop3_retr (arg);
      else if (strncasecmp (cmd, "DELE", 4) == 0)
	status = pop3_dele (arg);
      else if (strncasecmp (cmd, "NOOP", 4) == 0)
	status = pop3_noop (arg);
      else if (strncasecmp (cmd, "RSET", 4) == 0)
	status = pop3_rset (arg);
      else if ((strncasecmp (cmd, "TOP", 3) == 0) && (strlen (cmd) == 3))
	status = pop3_top (arg);
      else if (strncasecmp (cmd, "UIDL", 4) == 0)
	status = pop3_uidl (arg);
      else if (strncasecmp (cmd, "CAPA", 4) == 0)
	status = pop3_capa (arg);
      else
	status = ERR_BAD_CMD;

      if (status == OK)
	fflush (ofile);
      else if (status == ERR_WRONG_STATE)
	fprintf (ofile, "-ERR " BAD_STATE "\r\n");
      else if (status == ERR_BAD_ARGS)
	fprintf (ofile, "-ERR " BAD_ARGS "\r\n");
      else if (status == ERR_NO_MESG)
	fprintf (ofile, "-ERR " NO_MESG "\r\n");
      else if (status == ERR_NOT_IMPL)
	fprintf (ofile, "-ERR " NOT_IMPL "\r\n");
      else if (status == ERR_BAD_CMD)
	fprintf (ofile, "-ERR " BAD_COMMAND "\r\n");
      else if (status == ERR_BAD_LOGIN)
	fprintf (ofile, "-ERR " BAD_LOGIN "\r\n");
      else if (status == ERR_MBOX_LOCK)
	fprintf (ofile, "-ERR [IN-USE] " MBOX_LOCK "\r\n");
      else if (status == ERR_TOO_LONG)
	fprintf (ofile, "-ERR " TOO_LONG "\r\n");

      while (status == ERR_TOO_LONG)
	{
	  read (ifile, buf, 1);
	  if (buf[0] == '\n')
	    status = OK;
	}

      free (cmd);
      free (arg);
    }

  fflush (ofile);
  return OK;
}

/* Runs IDS POP in standalone daemon mode. This opens and binds to port 110,
   then executes a pop3_mainloop() upon accepting a connection. It starts 
   maxchildren child processes to listen to and accept socket connections */

int
pop3_daemon (int maxchildren)
{
  int children = 0;
  struct sockaddr_in client;
  int sock, sock2;
  unsigned int socksize;

  sock = socket (PF_INET, SOCK_STREAM, (getprotobyname ("tcp"))->p_proto);
  if (sock < 0)
    {
      syslog (LOG_ERR, "%m\n");
      exit (1);
    }
  memset (&client, 0, sizeof (struct sockaddr_in));
  client.sin_family = AF_INET;
  client.sin_port = htons (110);
  client.sin_addr.s_addr = htonl (INADDR_ANY);
  socksize = sizeof (client);
  if (bind (sock, (struct sockaddr *) &client, socksize))
    {
      perror ("Couldn't bind to socket");
      exit (1);
    }
  listen (sock, 128);
  while (1)
    {
      if (children < maxchildren)
	{
	  if (!fork ())
	    {
	      sock2 = accept (sock, &client, &socksize);
	      pop3_mainloop (sock2, sock2);
	      close (sock2);
	      exit (OK);
	    }
	  else
	    {
	      /* wait (NULL); */
	      children++;
	    }
	}
      else
	{
	  wait (NULL);
	  children--;
	}
    }
  return OK;
}
